/* odbcsdk-demo.c -- Simple demonstration of ODBCSDK
 * 091001
 *
 *                   Copyright August Software, USA, 1998 - 2002.
 *                Confidential & Proprietary.  All Rights Reserved.
 *
 *      ODBCSDK implements a cross-platform ODBC (driver) API that connects
 *      with data sources hosted on a centralized Windows NT/2K/XP system
 *      equipped with an ODBC ROUTER.  http://odbcrouter.com/
 *
 *                Module: ODBCSDK-DEMO  (ODBCSDK SDK Demo Application)
 *                   Rev: V1.3
 *                Source: MSVC6sp5 (X86) EGCS-1.1.2 (X86/PPC), etc..
 *              Synopsis: This module contains all of the source code
 *                        necessary to compile the ODBCSDK SDK's Demo App
 *
 *  HISTORY
 *  =============================================================
 *  ASC 090101 Initial Public Release from ODBCAPI on MacOS
 *  ASC 121101 Only WINDOWS.H on __WIN32__, lowercase sqlext.h
 *  ASC 122101 Don't use callback if USE_SYSTEM_CONFIG
 *  ASC 012502 make column index a short, update BIND_COLUMN macro
 */

/* Compile-time Options:
 *
 * define NUMERICS_ARE_DOUBLES to use 'doubles' instead of 'floats'
 *
 * define USE_SYSTEM_CONFIG to use /etc/odbcrouter.conf instead of
 * an application defined callback.  (note that an application defined
 * callback ALWAYS overrides /etc/odbcrouter.conf and that the new
 * /etc/odbcrouter.conf mechanism is available for Linux and Darwin
 * only, not for Win32, Classic Mac OS or Carbon Mac OS.)
 *
 * A multi-threaded version of this example is also available for
 * Linux, Darwin, Classic Mac OS and Carbon Mac OS, please contact
 * support@augsoft.com for a copy.
 */

/* file format for /etc/odbcrouter.conf (see also odbcsdk.h)
 *
 *     data_source_name = anet_address[, anet_directory,
 *                 (noprompt|prompt{=default_user_name})
 *                       [, connection_options]]
 *
 *    where data_source_name = "*|dsn"
 *      where * indicates applies to ALL dsns
 *            and dsn is an entry in SystemDSN tab on ODBCROUTER
 *
 *    where addr = "(A|I)#(GEOHUB|hostname|ip-address){:ip-port}"
 *      where A indicates AppleTalk, I indicates Internet IPv4
 *            and GEOHUB is the string GEOHUB
 *            and hostname is a DNS or AppleTalk hostname
 *            and ip-address is a dotted-quad IPv4 address
 *            and ip-port is an IPv4 port-num (req'd for I only)
 *
 *    where anet_directory = "(*|zonename|Internet)"
 *      where * indicates default directory/zone
 *            and zonename is an AppleTalk zonename
 *            and Internet is the string Internet
 *
 *    where connection_options (optional) = "KEY=VAL;KEY=VAL;..."
 *
 * examples:
 *
 * finance = A#GEOHUB, Finance & Administration, prompt=Bob
 * mrp = I#mrp.corp.googlecorp.com:6110, Internet, prompt
 * mkt = I#marketing:6110, Internet, noprompt, FILE=D:\MARKET.MDB
 * * = I#galaxia:6110, Internet, noprompt    <- catch all others
 *
 */

#define NUMERICS_ARE_DOUBLES
#ifndef __WIN32__
#define USE_SYSTEM_CONFIG
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __WIN32__
#include <WINDOWS.H>
#endif

#include <odbcsdk.h>  /* from AugSoft */

/* C macros
 */
#define SUCCESS_ODBC() ((rc == SQL_SUCCESS) || (rc == SQL_SUCCESS_WITH_INFO))
#define CHECK_ODBC( func ) rc = func; \
  if SUCCESS_ODBC() { printf( "o.k.\n" ); fflush( stdout ); } else break;
#define stringit(a) #a ""
#define BINDCOL(name,type) \
  printf( "Binding column [%s] (type=%d size=%d)..",stringit(name),type,sizeof(rec.name) ); fflush( stdout ); \
  rc = SQLBindCol(hstmt, i++, type, &rec.name, sizeof(rec.name), &cbrec.name); \
  if SUCCESS_ODBC() { printf( "ok.\n" ); fflush( stdout ); } else break;

/* These are some buffers to hold column (field) values in our table of records
 */
struct {
  long             AccountID;
  UCHAR            AccountName[256];
#ifdef NUMERICS_ARE_DOUBLES
  SDOUBLE          SomeDoubleType;
#else
  SFLOAT           SomeFloatType;
#endif
  TIMESTAMP_STRUCT SomeDateTimeType;
} rec = { 0, "", 0.0, {0, 0, 0, 0, 0, 0, 0} };

/* These are some buffers to hold returned field-lengths
 */
struct {
  SDWORD           AccountID;
  SDWORD           AccountName;
#ifdef NUMERICS_ARE_DOUBLES
  SDWORD           SomeDoubleType;
#else
  SDWORD           SomeFloatType;
#endif
  SDWORD           SomeDateTimeType;
} cbrec;

#ifndef USE_SYSTEM_CONFIG
/* Prototype
 *
 */
RETCODE FindMyServer(UCHAR*,UCHAR*,char*,char*,char*,char*,char*);

/* FindMyServer -- C callback for ODBCSDK to return the network-address
 *                 of an ODBC ROUTER serving any specified DSN (Data
 *                 Source Name), as well as any data source specific
 *                 connection parameters.  The return-code and arguments
 *                 are controlled by ODBCSDK, see ODBCSDK.H and the
 *                 SDK's "Quick Start" document for a thorough explanation.
 */

RETCODE FindMyServer(UCHAR *szDSN, UCHAR *gwId, char *szHostName,
                     char *szDirName, char *disAuthPrompt, char *szDefUser,
                     char *szConnOpts)
{
  if (szDSN)  /* for demo, we assume all DSNs are served by same ODBC ROUTER */
    {
      /* required, supply address of host ODBC ROUTER for this data source      */
      *gwId = 1;
      strcpy( szHostName, "I#galaxia:6110" ); /* use Internet to connect to <galaxia> on port 6110    */
      strcpy( szDirName, "*" );               /* zone-name only used for A#xxxx (AppleTalk) addresses */

      /* optional, additional information about this specific data-source       */
      *disAuthPrompt = 1;     /* don't display any user/password prompts        */
      szDefUser[0]   = '\0';  /* not needed, AuthPrompt will never be displayed */
      szConnOpts[0]  = '\0';  /* nothing to tack-on to outgoing connect-strings */

      return SQL_SUCCESS;
    }
  return SQL_ERROR;
}
#endif

int main(void)
{
  HENV             henv=(HENV)NULL;   /* Handle to entire ODBC Environment      */
  HDBC             hdbc=(HDBC)NULL;   /* Handle to specific data source         */
  HSTMT            hstmt=(HSTMT)NULL; /* Handle to specific SQL query/statement */
  RETCODE          rc = SQL_ERROR;    /* Convention for ODBC return-code        */

  UCHAR            szConnStrIn[256];  /* Our request to connect data-source     */
  UCHAR            szConnStrOut[256]; /* Data source's version of same request  */
  SWORD            cbLenStrOut = 0;   /* Returned length of szConnStrOut        */

  char             isConnected=0;     /* We'll set this to 1 when we are        */

  UCHAR            szSqlState[8];     /* Padded buffer for ODBC statecode       */
  UCHAR            szErrorMsg[SQL_MAX_MESSAGE_LENGTH]; /* ODBC message buffer   */
  SWORD            cbErrorMsg = 0;    /* Returned length of ODBC message        */
  SDWORD           fNativeError = 0;  /* Returned non-ODBC data source errcode  */
  short            i = 0;

  printf( "Welcome to ODBCSDK-Demo!\n\n" );
  srand(time(NULL));

#ifndef USE_SYSTEM_CONFIG
  ORSDK_InstallFindGWInfo( FindMyServer );  /* Install our callback rtn */
#endif

  printf( "------------------------------------------------------------\n" );

  do {

  /*
   * allocate the ODBC ENVIRONMENT (one per application) and
   * an ODBC CONNECTION (one per data source session)
   */



  printf( "Allocate the ODBC Environment..." ); fflush( stdout );
  printf( "Allocate the ODBC Environment..." ); fflush( stdout );
  CHECK_ODBC( SQLAllocEnv( &henv ) );
  printf( "------------------------------------------------------------\n" );
  printf( "Allocate a specific Connection handle..." ); fflush( stdout );
  CHECK_ODBC( SQLAllocConnect( henv, &hdbc ) );
  printf( "------------------------------------------------------------\n" );
  
  /* open the CONNECTION to a data source called LEDGER.  This data source must
   * be defined in the SystemDSN property-tab of the ODBC Control Panel of a 
   * Windows NT/2K/XP system somewhere in the network running our ODBC ROUTER.
   *
   * NOTE: ODBCAPI will callback to our handler (FindMyServer) passing it the
   * szDSN string "LEDGER" and expecting it to return the network address of
   * the corresponding ODBC ROUTER (from the Customer's preferences).
   */
  strcpy( (char *)szConnStrIn, "DSN=LEDGER;" );
  printf( "Attempting to access [%s]...", szConnStrIn ); fflush( stdout );
  CHECK_ODBC( SQLDriverConnect(hdbc, NULL, szConnStrIn,
                               (short)strlen( (char *)szConnStrIn ),
                               szConnStrOut, 255, &cbLenStrOut,
                               SQL_DRIVER_NOPROMPT) );
  isConnected = 1;
  printf( "Connected to [%s]\n", szConnStrOut ); fflush( stdout );
  /* a good application would now store the value returned in szConnStrOut
   * along with the data-source name in the Customer's preferences and
   * use it as the value of szConnStrIn the next time that it opens a
   * connection to this data source.  (note that this is not necessary,
   * but it may save the data source some extra work the next time around
   * and speed the connection process.)
   */
  printf( "------------------------------------------------------------\n" );
  
  /* allocate an ODBC STATEMENT (one per SQL query/statement)
   */
  printf( "Allocate Statement handle..." ); fflush( stdout );
  CHECK_ODBC( SQLAllocStmt( hdbc, &hstmt ) );
  printf( "------------------------------------------------------------\n" );
  
  /* execute a simple query to retrieve a variety of data types from a
   * a table in our data source called "Dreams".  note that we are reusing
   * a buffer we don't need anymore to pass the SQL statement.  you may
   * specify "generic ODBC SQL" which will work with ANY ODBC compliant
   * data source, or you may specify data source specific SQL (PL/SQL,
   * TransactSQL, etc.)
   */
  strcpy( (char *)szConnStrIn, "SELECT AccountID, AccountName, SomeDoubleType, SomeDateTimeType FROM Dreams;" );
  printf( "Attempting to execute [%s]...", szConnStrIn ); fflush( stdout );
  CHECK_ODBC( SQLExecDirect(hstmt, szConnStrIn, (SDWORD)strlen( (char *)szConnStrIn )) );
  printf( "------------------------------------------------------------\n" );
  
  /* bind our local variables to the columns (fields) within the table we are
   * querying.  note that the order in which these BINDCOL macros appear must
   * match the order that the columns appear in the SELECT statement.
   */
  i = 1;
  BINDCOL(AccountID, SQL_C_LONG);
  BINDCOL(AccountName, SQL_C_CHAR);
#ifdef NUMERICS_ARE_DOUBLES
  BINDCOL(SomeDoubleType, SQL_C_DOUBLE);
#else
  BINDCOL(SomeFloatType, SQL_C_FLOAT);
#endif
  BINDCOL(SomeDateTimeType, SQL_C_TIMESTAMP);
  printf( "------------------------------------------------------------\n" );
  
  /* now walk through the table of records, retrieving each record and
   * displaying its values.  note that "behind the scenes" ODBCAPI is
   * performing binary conversions between Intel and Motorola integer and
   * floating point formats.
   */
  while ( (rc = SQLFetch( hstmt )) != SQL_NO_DATA_FOUND )
    if SUCCESS_ODBC()
      {
        printf("AccountID=%5.5d\tAccountName=[%-32.32s]\tSomeDateTimeType=",
               rec.AccountID, rec.AccountName);
#ifdef NUMERICS_ARE_DOUBLES
        printf("%2.2d/%2.2d/%4.4d %2.2d:%2.2d:%2.2d.%4.4d, SomeDoubleType=%f\n",
#else
        printf("%2.2d/%2.2d/%4.4d %2.2d:%2.2d:%2.2d.%4.4d, SomeFloatType=%f\n",
#endif
               rec.SomeDateTimeType.month,
               rec.SomeDateTimeType.day,
               rec.SomeDateTimeType.year,
               rec.SomeDateTimeType.hour,
               rec.SomeDateTimeType.minute,
               rec.SomeDateTimeType.second,
               rec.SomeDateTimeType.fraction,
#ifdef NUMERICS_ARE_DOUBLES
               rec.SomeDoubleType
#else
               (double)rec.SomeFloatType
#endif
               );
      }
    else break;
  printf( "------------------------------------------------------------\n" );

  } while (0);
  printf( "\n============================================================\n" );

  /* ..retrieved results.  now if there are any "messages" queued up for us to
   * display for the user, we will display them here.  note that some kinds of
   * data sources, notably SQLServer and SYBASE, will queue-up "informational"
   * messages during a SQLConnect, SQLDriverConnect or SQLBrowseConnect.  When
   * this happens, these functions will return SQL_SUCCESS_WITH_INFO as an
   * indication that the Application is to retrieve these messages before
   * proceeding.  this demo, like early versions of FileMakerPro would interpret
   * that as an error and fall-through to this point.  we developed a work-around
   * for this problem in early FileMakerPro releases that can be activated by
   * specifying "II=1" in the connection options (specified in your callback or
   * in /etc/odbcrouter.com).  (II = "Ignore Informational" messages!)  this is
   * code shows how to purge the ODBC message-queue
   */
  do {
    rc = SQLError(henv, hdbc, hstmt, szSqlState, &fNativeError, szErrorMsg,
                  sizeof(szErrorMsg)-1, &cbErrorMsg);
    if ((rc == SQL_SUCCESS) || (rc == SQL_SUCCESS_WITH_INFO))
      {
        printf("\n***\nODBC ERROR: (%s) %s\nNATIVE ERR: (%ld)\n", szSqlState,
               szErrorMsg, fNativeError);
      }
  } while( rc == SQL_SUCCESS_WITH_INFO );
  
  printf( "\n============================================================\n" );

  /* now clean-up any remaining ODBC STATEMENT and CONNECTION handles we have
   * allocated and release the ODBC ENVIRONMENT.
   */
  if (hstmt)
    {
      printf( "Closing Statment handle\n" ); fflush( stdout );
      SQLFreeStmt( hstmt, SQL_CLOSE );
    }
  if (isConnected)
    {
      printf( "Disconnecting from Data Source\n" ); fflush( stdout );
      SQLDisconnect( hdbc );
    }
  if (hdbc)
    {
      printf( "Closing a specific Connection handle\n" ); fflush( stdout );
      SQLFreeConnect( hdbc );
    }
  if (henv)
    {
      printf( "Closing the ODBC Environment\n" ); fflush( stdout );
      SQLFreeEnv( henv );
    }

  printf( "Done.\n" );
}
